#!/usr/bin/env python

'''
'''

__docformat__ = 'restructuredtext'
__version__ = '$Id: $'

import math

import pyglet
from pyglet.gl import *
from pyglet import graphics

class BoneSprite(pyglet.sprite.Sprite):
    bone_dx = 0.
    bone_dy = 0.
    _bone_offset_rotation = 0.
    _bone_offset_rotation_flip_x = 0.
    _bone_size = 0.
    bone_parent = None

    _bone_next_x = 0.
    _bone_next_y = 0.
    _bone_next_rotation = 0.

    flip_x = False

    def __init__(self, *args, **kwargs):
        kwargs['blend_src'] = GL_ONE
        kwargs['blend_dest'] = GL_ONE_MINUS_SRC_ALPHA
        super(BoneSprite, self).__init__(*args, **kwargs)

    def __repr__(self):
        return self.name

    def init_bone(self, bone_dx, bone_dy):
        self.bone_dx = bone_dx
        self.bone_dy = bone_dy
        self._bone_size = math.sqrt(bone_dx * bone_dx + bone_dy * bone_dy)
        self._bone_offset_rotation = -math.atan2(bone_dy, bone_dx)
        self._bone_offset_rotation_flip_x = -math.atan2(bone_dy, -bone_dx)

    def _create_vertex_list(self):
        if self._batch is None:
            self._vertex_list = graphics.vertex_list(4,
                'v2f/%s' % self._usage, 
                'c4B', ('t3f', self._texture.tex_coords))
        else:
            self._vertex_list = self._batch.add(4, GL_QUADS, self._group,
                'v2f/%s' % self._usage, 
                'c4B', ('t3f', self._texture.tex_coords))
        self._update_position()
        self._update_color()

    def _update_position(self):
        flip_scale = -1 if self.flip_x else 1

        img = self._texture
        if not self._visible:
            self._vertex_list.vertices[:] = [0, 0, 0, 0, 0, 0, 0, 0]
            return

        x1 = -img.anchor_x * self._scale * flip_scale
        y1 = -img.anchor_y * self._scale
        x2 = x1 + img.width * self._scale * flip_scale
        y2 = y1 + img.height * self._scale

        parent = self.bone_parent
        if parent:
            x = parent._bone_next_x
            y = parent._bone_next_y

            rotation = self._rotation + parent._bone_next_rotation
        else:
            x = self._x
            y = self._y
            rotation = self._rotation
        self._bone_next_rotation = rotation

        r = rotation
        cr = math.cos(r)
        sr = math.sin(r)
        ax = x1 * cr - y1 * sr + x
        ay = x1 * sr + y1 * cr + y
        bx = x2 * cr - y1 * sr + x
        by = x2 * sr + y1 * cr + y
        cx = x2 * cr - y2 * sr + x
        cy = x2 * sr + y2 * cr + y
        dx = x1 * cr - y2 * sr + x
        dy = x1 * sr + y2 * cr + y

        bone_dx = self.bone_dx * flip_scale * self.scale
        bone_dy = self.bone_dy * self.scale
        self._bone_next_x = bone_dx * cr - bone_dy * sr + x
        self._bone_next_y = bone_dx * sr + bone_dy * cr + y

        self._vertex_list.vertices[:] = [ax, ay, bx, by, cx, cy, dx, dy]

    def get_position(self):
        parent = self.bone_parent
        if parent:
            return parent._bone_next_x, parent._bone_next_y
        return self._x, self._y

    def get_bone_offset_rotation(self):
        if self.flip_x:
            return self._bone_offset_rotation_flip_x
        else:
            return self._bone_offset_rotation

    def get_offset_rotation_world(self):
        '''world-space rotation to x axis'''
        parent = self.bone_parent
        if parent:
            return self.get_bone_offset_rotation() - parent._bone_next_rotation
        else:
            return self.get_bone_offset_rotation()

    def get_offset_rotation_local(self):
        parent = self.bone_parent
        if parent:
            return self.get_bone_offset_rotation() - parent.get_bone_offset_rotation()
        else:
            return self.get_bone_offset_rotation()

    def interp_rotation(self, target, speed):
        self._rotation += (target - self._rotation) * speed

    def ik(self, x3, y3, flip_knee=False, speed=1):
        flip_knee ^= self.flip_x

        parent = self.bone_parent
        if parent:
            x1, y1 = parent.get_position()
        else:
            x1, y1 = self.get_position()

        # Vector connecting parent root to target
        base_dx = x3 - x1
        base_dy = y3 - y1
        base_rotation = math.atan2(base_dy, base_dx)

        if parent:
            # Find distance to pull knee in by
            base_size = math.sqrt(base_dx * base_dx + base_dy * base_dy)
            bone_size = self._bone_size + parent._bone_size

            if base_size >= bone_size:
                # Fully stretched out
                parent.interp_rotation(base_rotation + parent.get_offset_rotation_world(), speed)
                self.interp_rotation(self.get_offset_rotation_local(), speed)
                return False
            else:
                # Solve for two bones
                knee = math.acos(base_size / bone_size)
                if flip_knee:
                    knee = -knee
                parent.interp_rotation(knee +
                    base_rotation + parent.get_offset_rotation_world(), speed)
                self.interp_rotation(-2 * knee + 
                    self.get_offset_rotation_local(), speed)

        else:
            # Single bone chain
            self.interp_rotation(base_rotation + self.get_offset_rotation_world(), speed)
            
        return True
